/*****************************************************************************
 * Filename:    cryptfs/options.c                                            *
 * Description: Mangement of the configuration information                   *
 * Copyright:   2005 by Alexander Motzkau                                    *
 *****************************************************************************/
#define _GNU_SOURCE
#include <search.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "misc.h"
#include "options.h"

int compare_action(const void* a, const void* b)
{
    return strcmp(((struct action*)a)->name, ((struct action*)b)->name);
}

int compare_key(const void* a, const void* b)
{
    return strcmp(((struct key*)a)->id, ((struct key*)b)->id);
}

int compare_storage(const void* a, const void* b)
{
    return strcmp(((struct storage_block*)a)->entry->name,
		  ((struct storage_block*)b)->entry->name);
}

enum bool complete_config(struct configuration* config)
{
    struct action* ac;
    struct storage* st;
    
    config->action_tree = NULL;
    config->storage_tree = NULL;
    config->key_tree = NULL;

    for(st = config->storage; st != NULL; st = st->next)
    {
	struct storage_entry* entry;
	
	for(entry = st->entries; entry != NULL; entry = entry->next)
	{
	    struct storage_block** ste_val;
	    struct storage_block* tree_ent;

	    tree_ent = (struct storage_block*) xmalloc(sizeof(struct storage_block));
	    tree_ent->device = st;
	    tree_ent->entry = entry;
	    tree_ent->next = NULL;
	    
	    ste_val = (struct storage_block**) tsearch(tree_ent, &config->storage_tree, compare_storage);
	    if(ste_val == NULL)
	    {
		errmsg("Out of memory.\n");
		xfree(tree_ent);
		return false;
	    }
	    
	    if(*ste_val != tree_ent)
	    {
		struct storage_block* temp;
		
		for(temp = *ste_val; temp->next != NULL; temp = temp->next);
		    /* Do nothing. */
		    
		temp->next = tree_ent;
	    }
	}
    }
    
    for(ac = config->actions; ac != NULL; ac = ac->next)
    {
	struct action** ac_val;
	struct key** k_val;
	enum bool have_key;
	struct map* mapping;

	ac_val = (struct action**) tsearch(ac, &config->action_tree, compare_action);
	if(ac_val == NULL)
	{
	    errmsg("Out of memory.\n");
	    return false;
	}
	    
	if(*ac_val != ac)
	{
	    errmsg("action '%s' occurs more than once.\n", ac->name);
	    return false;
	}
	
	if(ac->key == NULL)
	    have_key = false;
	else
	{
	    if(ac->key->id != NULL)
	    {
		k_val = (struct key**) tsearch(ac->key, &config->key_tree, compare_key);
	    
		if(k_val == NULL)
		{
		    errmsg("Out of memory.\n");
	    	    return false;
		}

		if(*k_val == ac->key && ac->key->type == ref)
		{
		    errmsg("Unknown referenced key '%s'.\n", ac->key->id);
		    return false;
		}
	    
		/* The first entry with this id counts. */
	    }
	    
	    have_key = true;
	}
	    
	for(mapping = ac->mappings; mapping != NULL; mapping = mapping->next)
	{
	    if(mapping->key != NULL)
	    {
		if(mapping->key->id != NULL)
		{
		    k_val = (struct key**) tsearch(mapping->key, &config->key_tree, compare_key);
		    
		    if(k_val == NULL)
		    {
			errmsg("Out of memory.\n");
			return false;
		    }
		}
		
	    }
	    else if(!have_key)
	    {
		errmsg("No key given in action '%s'.\n", ac->name);
		return false;
	    }
	    
	    if(find_storage(mapping->name, config)==NULL)
	    {
		errmsg("Unknown device '%s' to be mapped.\n", mapping->name);
		return false;
	    }
	}
    }
    
    return true;
}

struct action* find_action(char *name, struct configuration* config)
{
    struct action** val;
    struct action ac;

    ac.name = name;
    val = (struct action**) tfind(&ac, &config->action_tree, compare_action);
    if(val == NULL)
	return NULL;
	
    return *val;
}

struct storage_block* find_storage(char *name, struct configuration* config)
{
    struct storage_block** val;
    struct storage_entry entry;
    struct storage_block tree_ent;

    tree_ent.entry = &entry;
    entry.name = name;
    
    val = (struct storage_block**) tfind(&tree_ent, &config->storage_tree, compare_storage);

    return *val;
}

struct key* find_key(char *name, struct configuration* config)
{
    struct key** val;
    struct key k;

    k.id = name;
    val = (struct key**) tfind(&k, &config->key_tree, compare_key);
    if(val == NULL)
	return NULL;
	
    return *val;
}

#define REMOVE(type, list, fun)			\
    do						\
    {						\
	type *R_entry;				\
	R_entry = list;				\
						\
	while(R_entry != NULL)			\
	{					\
	    type *R_temp;			\
						\
	    R_temp = R_entry;			\
	    R_entry = R_entry->next;		\
	    fun(R_temp);			\
	}					\
    } while(0)

void free_entry(struct storage_entry* entry)
{
    xfree(entry->name);
    xfree(entry->cipher);
    xfree(entry);
}

void free_storage(struct storage* storage)
{
    xfree(storage->device);
    REMOVE(struct storage_entry, storage->entries, free_entry);
    xfree(storage);
}

void free_key(struct key *key)
{
    xfree(key->id);
    xfree(key->algo);
    xfree(key->value);
    xfree(key->uid);
    
    REMOVE(struct key, key->subkeys, free_key);
    
    if(key->result != NULL && key->type != ref)
	xfree(key->result);

    xfree(key);
}

void free_command(struct command *cmd)
{
    xfree(cmd->uid);
    xfree(cmd->cmd);
    xfree(cmd);
}

void free_map(struct map *map)
{
    xfree(map->name);
    xfree(map->point);
    xfree(map->fs);
    xfree(map->options);

    REMOVE(struct command, map->preexec, free_command);
    REMOVE(struct command, map->postexec, free_command);

    if(map->key != NULL) free_key(map->key);
    
    xfree(map);
}

void free_action(struct action* act)
{
    xfree(act->name);
    REMOVE(struct command, act->preexec, free_command);
    REMOVE(struct command, act->postexec, free_command);
    REMOVE(struct map, act->mappings, free_map);
    if(act->key != NULL) free_key(act->key);
    xfree(act);
}

void free_dummy(void *node) {}

void free_config(struct configuration *config)
{
    tdestroy(config->action_tree, free_dummy);
    tdestroy(config->key_tree, free_dummy);
    tdestroy(config->storage_tree, xfree);
    REMOVE(struct action, config->actions, free_action);
    REMOVE(struct storage, config->storage, free_storage);
    xfree(config);
}
